@@ -206,7 +206,9 @@ MPL_STATIC_INLINE_PREFIX int MPIDIG_mpi_win_complete(MPIR_Win * win)
     for (win_grp_idx = 0; win_grp_idx < group->size; ++win_grp_idx) {
         peer = ranks_in_win_grp[win_grp_idx];
 
-        CH4_CALL(am_send_hdr(peer, win->comm_ptr, MPIDIG_WIN_COMPLETE, &msg, sizeof(msg), vci, vci),
+        int vci_target = MPIDI_WIN_TARGET_VCI(win, peer);
+        CH4_CALL(am_send_hdr
+                 (peer, win->comm_ptr, MPIDIG_WIN_COMPLETE, &msg, sizeof(msg), vci, vci_target),
                  MPIDI_rank_is_local(peer, win->comm_ptr), mpi_errno);
         if (mpi_errno != MPI_SUCCESS)
             MPIR_ERR_SETANDSTMT(mpi_errno, MPI_ERR_RMA_SYNC, goto fn_fail, "**rmasync");
@@ -265,8 +267,9 @@ MPL_STATIC_INLINE_PREFIX int MPIDIG_mpi_win_post(MPIR_Group * group, int assert,
 
     for (win_grp_idx = 0; win_grp_idx < group->size; ++win_grp_idx) {
         peer = ranks_in_win_grp[win_grp_idx];
-
-        CH4_CALL(am_send_hdr(peer, win->comm_ptr, MPIDIG_WIN_POST, &msg, sizeof(msg), vci, vci),
+        int vci_target = MPIDI_WIN_TARGET_VCI(win, peer);
+        CH4_CALL(am_send_hdr
+                 (peer, win->comm_ptr, MPIDIG_WIN_POST, &msg, sizeof(msg), vci, vci_target),
                  MPIDI_rank_is_local(peer, win->comm_ptr), mpi_errno);
         if (mpi_errno != MPI_SUCCESS)
             MPIR_ERR_SETANDSTMT(mpi_errno, MPI_ERR_RMA_SYNC, goto fn_fail, "**rmasync");
@@ -354,6 +357,7 @@ MPL_STATIC_INLINE_PREFIX int MPIDIG_mpi_win_lock(int lock_type, int rank, int as
 
     MPIR_FUNC_ENTER;
     int vci = MPIDI_WIN(win, am_vci);
+    int vci_target = MPIDI_WIN_TARGET_VCI(win, rank);
     MPID_THREAD_CS_ENTER(VCI, MPIDI_VCI(vci).lock);
 
     MPIDIG_LOCK_EPOCH_CHECK_NONE(win, rank, mpi_errno, goto fn_fail);
@@ -374,7 +378,7 @@ MPL_STATIC_INLINE_PREFIX int MPIDIG_mpi_win_lock(int lock_type, int rank, int as
     msg.lock_type = lock_type;
 
     locked = slock->locked + 1;
-    CH4_CALL(am_send_hdr(rank, win->comm_ptr, MPIDIG_WIN_LOCK, &msg, sizeof(msg), vci, vci),
+    CH4_CALL(am_send_hdr(rank, win->comm_ptr, MPIDIG_WIN_LOCK, &msg, sizeof(msg), vci, vci_target),
              MPIDI_rank_is_local(rank, win->comm_ptr), mpi_errno);
 
     if (mpi_errno != MPI_SUCCESS)
@@ -427,6 +431,7 @@ MPL_STATIC_INLINE_PREFIX int MPIDIG_mpi_win_unlock(int rank, MPIR_Win * win)
 #endif
 
     int vci = MPIDI_WIN(win, am_vci);
+    int vci_target = MPIDI_WIN_TARGET_VCI(win, rank);
     MPID_THREAD_CS_ENTER(VCI, MPIDI_VCI(vci).lock);
     need_unlock = 1;
 
@@ -443,7 +448,8 @@ MPL_STATIC_INLINE_PREFIX int MPIDIG_mpi_win_unlock(int rank, MPIR_Win * win)
     msg.origin_rank = win->comm_ptr->rank;
     unlocked = slock->locked - 1;
 
-    CH4_CALL(am_send_hdr(rank, win->comm_ptr, MPIDIG_WIN_UNLOCK, &msg, sizeof(msg), vci, vci),
+    CH4_CALL(am_send_hdr
+             (rank, win->comm_ptr, MPIDIG_WIN_UNLOCK, &msg, sizeof(msg), vci, vci_target),
              MPIDI_rank_is_local(rank, win->comm_ptr), mpi_errno);
     if (mpi_errno != MPI_SUCCESS)
         MPIR_ERR_SETANDSTMT(mpi_errno, MPI_ERR_RMA_SYNC, goto fn_fail, "**rmasync");
@@ -700,7 +706,9 @@ MPL_STATIC_INLINE_PREFIX int MPIDIG_mpi_win_unlock_all(MPIR_Win * win)
         msg.win_id = MPIDIG_WIN(win, win_id);
         msg.origin_rank = win->comm_ptr->rank;
 
-        CH4_CALL(am_send_hdr(i, win->comm_ptr, MPIDIG_WIN_UNLOCKALL, &msg, sizeof(msg), vci, vci),
+        int vci_target = MPIDI_WIN_TARGET_VCI(win, i);
+        CH4_CALL(am_send_hdr
+                 (i, win->comm_ptr, MPIDIG_WIN_UNLOCKALL, &msg, sizeof(msg), vci, vci_target),
                  MPIDI_rank_is_local(i, win->comm_ptr), mpi_errno);
         if (mpi_errno != MPI_SUCCESS)
             MPIR_ERR_SETANDSTMT(mpi_errno, MPI_ERR_RMA_SYNC, goto fn_fail, "**rmasync");
@@ -851,7 +859,9 @@ MPL_STATIC_INLINE_PREFIX int MPIDIG_mpi_win_lock_all(int assert, MPIR_Win * win)
         msg.origin_rank = win->comm_ptr->rank;
         msg.lock_type = MPI_LOCK_SHARED;
 
-        CH4_CALL(am_send_hdr(i, win->comm_ptr, MPIDIG_WIN_LOCKALL, &msg, sizeof(msg), vci, vci),
+        int vci_target = MPIDI_WIN_TARGET_VCI(win, i);
+        CH4_CALL(am_send_hdr
+                 (i, win->comm_ptr, MPIDIG_WIN_LOCKALL, &msg, sizeof(msg), vci, vci_target),
                  MPIDI_rank_is_local(i, win->comm_ptr), mpi_errno);
         if (mpi_errno != MPI_SUCCESS)
             MPIR_ERR_SETANDSTMT(mpi_errno, MPI_ERR_RMA_SYNC, goto fn_fail, "**rmasync");
