@@ -1,11 +1,9 @@
 ! This file created from f77/rma/c2f2cwinf.f with f77tof90
-! -*- Mode: Fortran; -*- 
 !
-!  (C) 2003 by Argonne National Laboratory.
-!      See COPYRIGHT in top-level directory.
-!   
-! Test just MPI-RMA
+! Copyright (C) by Argonne National Laboratory
+!     See COPYRIGHT in top-level directory
 !
+
       program main
       use mpi
       integer errs, toterrs, ierr
@@ -18,7 +16,7 @@
 
       errs = 0
 
-      call mpi_init( ierr )
+      call mtest_init( ierr )
 
 !
 ! Test passing a Fortran MPI object to C
@@ -39,15 +37,7 @@
 !
 ! Summarize the errors
 !
-      call mpi_allreduce( errs, toterrs, 1, MPI_INTEGER, MPI_SUM, &
-      &     MPI_COMM_WORLD, ierr )
-      if (wrank .eq. 0) then
-         if (toterrs .eq. 0) then
-            print *, ' No Errors'
-         else
-            print *, ' Found ', toterrs, ' errors'
-         endif
-      endif
+      call mtest_finalize( errs )
 
       end
       
